
var Controller = new Phaser.Class({
	Extends: Phaser.Scene,

    initialize: function Controller (){
        Phaser.Scene.call(this, 'Controller');
    },

    init: function (){
		this.cameras.main        
            .setSize(650, 450);	

        this.input.on('pointerdown', () => {
            clickOffCanvas = false;
        });
	},

    preload: function (){

    },

    create: function (){
        this.setup();

        this.registry.values.uiState = {
            textbox: false,
            sorting: false,

            menuState: undefined,

            get uiIsActive(){
                return (this.textbox || this.menuState || this.sorting)
            }
        }

        this.registry.values.items = {
            sword: {
                icon: "item_sword",
                tooltip: "tooltip_sword",
            },
            lantern: {
                icon: "item_lantern",
                icon_anim: "item_lantern_anim",
                tooltip: "tooltip_lantern",
                tooltip_anim: "tooltip_lantern_anim",
            },
            note: {
                icon: "item_note",
                tooltip: "tooltip_note",
            },
            healing_hands: {
                icon: "item_healing_hands",
                tooltip: "tooltip_healing_hands",
            },
            flower_crown: {
                icon: "item_flower_crown",
                tooltip: "tooltip_flower_crown",
            },
        }
        
        //globalVolume should determine the volume of everything in the game
        //Make sure every sound gets set to globalVolume where relevant, to ensure nothing plays while the game is muted.
        this.registry.values.globalVolume = 1;
        this.registry.values.muteStage = 0;
        this.registry.values.music = undefined;
        this.registry.values.tracks = {};
        this.registry.values.tracks.bgm = this.sound.add("bgm", {loop: true});
        this.registry.values.ambient = this.sound.addAudioSprite("ambient");
        this.registry.values.keys = this.input.keyboard.addKeys('W,A,S,D,E,SPACE');   
   
        this.registry.values.transitioning = false;         

        this.scene.add("Forest", Forest)
        this.scene.add("Menu", Menu)
        this.scene.add("Tooltip", Tooltip)
        this.scene.add("Textbox", Textbox)
        this.scene.add("Text", Text)
        this.scene.add("Volume", Volume)

        this.scene.launch("Textbox")
        this.scene.launch("Text")
        this.scene.launch("Menu")
        this.scene.launch("Volume")
        
        this.registry.values.activeArea = "Forest";
        this.scene.launch("Forest");	

        this.scene.get("Volume").playTrack("bgm");
    },

    update: function (){
        if (this.data.values.pendingReset && !this.scene.get("Forest")) {
            this.scene.add("Forest", Forest)

            this.registry.values.activeArea = "Forest";
            this.scene.launch("Forest")

            this.scene.get("Volume").playTrack("bgm");
        }
    },    

    disableMenu: function(){
        this.scene.get("Menu").disable();
    },

    enableMenu: function(){
        this.scene.get("Menu").enable();
    },

    //Putrid fucking hack for an extremely short sighted use-case. If inventory comes back in a big way, nuke all of this
    inventoryUpdate: function(chara){
        let forestData = this.scene.get("Forest").data.values;
        if (chara.inv[0] == "lantern" || chara.inv[1] == "lantern") {
            this.registry.values.safe = true;

            forestData.light.setVisible(true);
            forestData.overlay.setMask(forestData.mask);
            this.registry.values.cDatas.pData.textureKey = "chara_nino";

            this.scene.get("Volume").setMusicVolume(1, 500)
        }
        else {
            this.registry.values.safe = false;

            forestData.overlay.clearMask();
            forestData.light.setVisible(false);
            this.registry.values.cDatas.pData.textureKey = "nino_dark";
            this.scene.get("Volume").setMusicVolume(0, 20000)
        }
    },
    
    addToInventory: function(chara, item) {
        if (!(item in this.registry.values.items)) { console.warn(`Item ${item} isn't real`) }

        this.registry.values.party.forEach(member => {
            if (member.chara === chara) {
                for (let i = 2; i < 6; i++) {
                    if (!member.inv[i]) {
                        member.inv[i] = item;
                        return;
                    }
                }
            }
        });

    },

    setup: function() {
        this.registry.values.flags = new Set();
        this.registry.values.safe = true;

        this.registry.values.party = [
            {
                chara: "nino",
                colour: 0x008000,
                tooltip: "tooltip_nino_hidden",
                inv: [ "sword", "lantern", "note", "healing_hands", undefined, "flower_crown",  ]
            }
        ]

    },

    restartGame: function (){

        this.scene.get("Volume").stopMusic();
        this.registry.values.activeArea = undefined;

        ["Forest"].forEach((deadScene) => {
            this.scene.get(deadScene).scene.remove();
        });

        this.setup();

        this.registry.values.cDatas.pData = {
            key: "player",
            textureKey: "chara_nino",
            spawnKey: "spawn",
            initialDirection: "left",
            sizeX: 22, 
            sizeY: 12,
            offsetX: 6,
            offsetY: 36,
            speed: 175,
            stillframes: {down:0, left:3, right:6, up:9}
        };
        
        this.data.values.pendingReset = true;
    }
});

var Forest = new Phaser.Class({ 
	Extends: Phaser.Scene,

	initialize: function Forest(){
	    Phaser.Scene.call(this, 'Forest');
	},

	init: function (){
		this.input.on('pointerdown', () => {
		    clickOffCanvas = false;
		});
	},

	preload: function (){
	},

	create: function(){ 
		this.scene.sendToBack();
		
		//Declare scene constants	
		let gv = this.registry.values;	
		let sv = this.data.values;

		//Add the environment
		sv.backdrop = this.add
			.image(0, 0, 'forestBG')
			.setOrigin(0);

		sv.light = this.add
			.image(0,0,'atlas', 'mask')
			.setScale(2)
			.setAlpha(0.3)
			.setBlendMode('ADD');
		sv.mask = sv.light.createBitmapMask();
		sv.mask.invertAlpha = true;

		sv.overlay = this.add
			.rectangle(0, 0, 650, 450, 0x000000, 0.92)
			.setScrollFactor(0)
			.setOrigin(0)
			.setDepth(50)
			.setMask(sv.mask)

		//Create map
		sv.map = this.make.tilemap({key: "forestMap"});

		//Get all objects from map
		sv.flags = this.physics.add.group();
		sv.interacts = this.physics.add.group();
		sv.flags.addMultiple(objectArrayToImageArray(sv.map.getObjectLayer("FlagLayer").objects, this));
		sv.interacts.addMultiple(objectArrayToImageArray(sv.map.getObjectLayer("InteractLayer").objects, this));

		sv.walls = this.physics.add.staticGroup().addMultiple(objectArrayToImageArray(sv.map.getObjectLayer("WallLayer").objects, this));

		//Sets up camera
		this.cameras.main
			.setSize(650, 450)
			.setBounds(0, 0, 1520, 2176);

		//Array to keep track of which flags the player is touching
		sv.isOverlappingFlag = false;
		sv.currentFlags = [];

		//Create the player
		sv.actors = new gameActorList(this);
		sv.actors.newPlayer(gv.cDatas.pData);
		sv.actors.newCutsceneActor({
			key: "dummy",
			textureKey: "chara_nino",
			spawnKey: "spawn",
			stillframes: {down:0, left:3, right:6, up:9}
		});
		sv.actors.objects['dummy'].setVisible(false);

		this.physics.world.addCollider(sv.actors.player, sv.walls);
		
		this.tweens.add({
			targets: this.data.values.light,
			scaleX: 1.9,
			scaleY: 1.9,
			duration: 100,
			yoyo: true,
			loop: -1
		});

		sv.actors.push('resetButton', this.physics.add
			.image(10, 440, 'atlas', 'ui_back')
			.setScrollFactor(0)
			.setDepth(55)
            .setOrigin(0, 1)
			.setScale(2)
			.setAlpha(0)
            .on('pointerover', event => {  
                this.tweens.add({
                    targets: this.data.values.actors.objects.resetButton,
                    alpha: 0.8,
                    duration: 60
                });
            })    
            .on('pointerout', event => {    
                this.tweens.add({
                    targets: this.data.values.actors.objects.resetButton,
                    alpha: 0.4,
                    duration: 60
                });    
            })    
            .on('pointerdown', () => {
                this.cameras.main.once('camerafadeoutcomplete', event => {
					this.scene.stop();
					this.scene.get("Controller").restartGame();
				});	
				this.registry.values.ambient.off("complete");
                this.cameras.main.fadeOut(250);
			})
		,);

		this.registry.values.transitioning = false;
		this.cameras.main.fadeIn(200);

		this.time.delayedCall(50, ()=>{
			callTextbox("startText")
		})

		this.events.on('postupdate', () => {
			let target = (this.data.values.targetDummy) ? this.data.values.actors.objects.dummy : this.data.values.actors.objects.player;
			this.data.values.light.setPosition(target.x, target.y);
		})

		this.registry.values.ambient.on("complete", ()=> {
			let time = 5000;
			let vol = 1;
			if (this.data.values.longend){
				time = 200;
				vol = 1 + 2*(this.data.values.actors.player.y / 2176) + (this.registry.values.safe ? 0 : 2)
			}
			else if (this.registry.values.safe) {
				time += Math.floor(Math.random()*5000)
				vol = 0.5 + Math.random();
			}
			else {
				time -=  (4000 - this.registry.values.music.volume * 4000) + Math.floor(Math.random()*1000)
				vol = 5.3 - 5*this.registry.values.music.volume;
			}
			this.time.delayedCall(time, ()=> {
				if (!this.data.values.stopAmbient) {
					if (this.data.values.howlPending) {
						this.data.values.howlPending = false;
						let i = Math.floor(Math.random()*5)+1
						vol = 3 - 2.6*(this.data.values.actors.player.y / 2176)
						this.registry.values.ambient.play(`howl${i}`, {volume: vol})
					}
					else {
						let i = Math.floor(Math.random()*15)+1
						this.registry.values.ambient.play(`crunch${i}`, {volume: vol})
					}
				}
			})
		})
		this.time.delayedCall(1000, ()=> {
			this.registry.values.ambient.play(`howl4`, {volume: 1})
		})
	},

	update: function(time, delta) {
		

		//Checks for examinable objects when you smash that subscribe button
		//Not when there's a dialogue box active though. Spacebar has more important things to do then
		if (!this.registry.values.uiState.uiIsActive && Phaser.Input.Keyboard.JustDown(this.registry.values.keys.SPACE)){
			//Decide whether player is on lower portion of screen
			setTextboxTop((this.data.values.actors.player.y - this.cameras.main.scrollY) > 265);

			let isEnvironmentalInteract = this.physics.overlap(this.data.values.pInteractBox, this.data.values.interacts, this.interactHandler, null, this);
			if (!isEnvironmentalInteract)
				this.physics.overlap(this.data.values.pInteractBox, this.data.values.actors.npcs, (interactbox, npc) =>{
					this.interactHandler(interactbox, npc)
					npc.interactHandler(this.data.values.actors.player);
				}, null, this);
		}

		//Unset flags that player is no longer touching
		if (this.data.values.currentFlags.length > 0){			
			for (let i = 0; i < this.data.values.currentFlags.length; i++){
				let isOverlappingFlag = false;	
				this.physics.overlap(this.data.values.actors.player, this.data.values.currentFlags[i], ()=>{isOverlappingFlag = true;});
				if (!isOverlappingFlag) this.data.values.currentFlags.splice(i, 1);
			}
		}

		this.data.values.actors.update();
	},
	
	interactHandler: function(interactBox, interact){
		//Handle environment interaction
		switch (interact.name){
			case "interact_tulip1":
				if (this.registry.values.party[0].inv[0] == "lantern" || this.registry.values.party[0].inv[1] == "lantern")
					callTextbox("tulip1");
				else
					callTextbox("darkTulip");
				break;
			case "interact_tulip2":
				if (this.registry.values.party[0].inv[0] == "lantern" || this.registry.values.party[0].inv[1] == "lantern")
					callTextbox("tulip2");
				else
					callTextbox("darkTulip");
				break;		

			case "interact_directions":
				if (this.registry.values.party[0].inv[0] == "lantern" || this.registry.values.party[0].inv[1] == "lantern")
					callTextbox("directions");
				else
					callTextbox("darkSign");
				break;
			case "interact_redinn":
				if (this.registry.values.party[0].inv[0] == "lantern" || this.registry.values.party[0].inv[1] == "lantern") {	
					if (!this.registry.values.flags.has('read_redinn'))
						callTextbox("redinn");
					else 
						callTextbox('redinn_repeat')
				}
				else
					callTextbox("darkSign");
				break;	
		}
	},

	flagHandler: function(player, flag){
		let lock = false;
		for (let i = 0; i < this.data.values.currentFlags.length; i++){
			if (this.data.values.currentFlags[i].name == flag.name){
				lock = true;
			}
		}
		
		if (!lock){
			this.data.values.currentFlags.push(flag);
			switch (flag.name){
				
				case "flag_startend":
					callTextbox("startend")
					break;

				case "flag_deadend":
					flag.destroy();
					if (!this.registry.values.flags.has('read_deadend1'))
						callTextbox("deadend1");
					else if (!this.registry.values.flags.has('read_deadend2'))
						callTextbox("deadend2");
					else if (!this.registry.values.flags.has('read_deadend3'))
						callTextbox("deadend3");
					else 
						callTextbox("deadend4");
					break;

				case "flag_longend":
					if (this.registry.values.safe && !this.registry.values.flags.has("read_longend"))
						callTextbox("longend")
					else {
						this.cameras.main.once('camerafadeoutcomplete', event => {
							this.scene.stop();
							this.scene.get("Controller").time.delayedCall(2000, ()=>{
								this.scene.get("Controller").restartGame();
							})
						});	
						this.registry.values.ambient.off("complete");
						this.registry.values.ambient.stop();
						this.cameras.main.fadeOut(1);
						this.scene.get("Volume").stopMusic();
					}

					break;

				case "flag_endend":
					callTextbox("endend")
					break;

				case "flag_howl":
					flag.destroy();
					this.data.values.howlPending = true;
					break;
				case "flag_safe":
					this.data.values.longend = false;
					break;
				case "flag_crunch":
					this.registry.values.ambient.stop();
					let i = Math.floor(Math.random()*15)+1
					this.registry.values.ambient.play(`crunch${i}`, {volume: 3})
					this.data.values.howlPending = false;
					this.data.values.longend = true;
					break;
			}
		}
	},

	wakeCallback: function() {
	},

	sleepCallback: function() {
		
	},
});

var globalScript =  [{
    startText:{
        text: [
			"You can hear a howling noise...\nYou need to get out of here fast!",
			"Move with the WASD keys, \nand examine your surroundings \nwith the SPACE bar. Press E to \nopen your inventory."
        ]
    },

    tulip1:{
        text: [
            "Red tulips.",
            "Looks unhealthy."
        ]
    },
    tulip2:{
        text: [
            "Red tulips.",
            "Smells like blood."
        ]
    },
    darkTulip:{
        text: [
            "Something here smells bloody."
        ]
    },


	darkSign:{
		text: [
			"You can't read the sign without\nlight."
		]	
	},
    redinn:{
        text: [
            "It's a sign.",
            "It reads... Visit Red's Inn.\nA great stay is a guarantee.",
            "It's so comfortable, you will not\nwant to leave. ",
            "Ever.",
            "Visit Red's Inn.",
            "...",
            "You don't think you'll visit\nRed's Inn."
        ]
    },
    redinn_repeat:{
        text: [
            "...You don't think you'll visit\nRed's Inn.",
        ]
    },

    directions:{
        text: [
            "A sign for directions.",
            "RIGHT: Dead end.\nLEFT: Dead end.\nDOWN: Dead end, probably.\nUPWARDS: Red Tulip Town."
        ]
    },

    deadend1:{
        text: [
            "It's a dead end."
        ]
    },
    deadend2:{
        text: [
            "Darn, a dead end?"
        ]
    },
    deadend3:{
        text: [
            "Nothing to see here, just a dead\nend."
        ]
    },
    deadend4:{
        text: [
            "So many dead ends?????"
        ]
    },

    longend:{
        text: [
			()=>"step1",
            "All that trouble for a dead end?\nThis is really tiring you out."
		],
		step1(){
			setTextboxTop(true)
			callTextbox();
		}
    },

    startend:{
        text: [
			"This is the path you came from!\nYou're not going back there.",
			()=>"step1"
        ],
        step1(controller) {
			let scene = controller.scene.get("Forest");
			let dummy = scene.data.values.actors.objects.dummy;
			let player = scene.data.values.actors.player;

			controller.registry.values.cutscene = true;
			scene.data.values.targetDummy = true;
            hideActiveTextbox();

			dummy.swapIn(player);
			scene.cameras.main.startFollow(dummy);
			dummy.moveTo(dummy.x - 50, dummy.y, 200, 750);

			controller.time.delayedCall(750, ()=> {
                dummy.swapOut(player)
                scene.cameras.main.startFollow(player)
				scene.data.values.targetDummy = false;
            })
            
            controller.time.delayedCall(850, ()=> {
                stopActiveTextbox();
                controller.registry.values.cutscene = false;
				callTextbox();
			})
			
        }
    },

    endend:{
        text: [
            "The howling is coming from this\ndirection. This is the right way.",
            [
                ["This will end the gam&[2]e. C&[3]ontinue?"],
                ["Yes.", "ending"],
                ["No.", "cancelEnding"]
            ]
        ]
    },
    ending: {
        text: [
            ()=>"step1"
        ],
        step1(controller) {
			
			

			let scene = controller.scene.get("Forest");
			let dummy = scene.data.values.actors.objects.dummy;
			let player = scene.data.values.actors.player;
			
			controller.registry.values.cutscene = true;
			scene.data.values.targetDummy = true;
			scene.data.values.stopAmbient = true;
            hideActiveTextbox();

			dummy.swapIn(player);
			scene.cameras.main.startFollow(dummy);
			dummy.moveTo(dummy.x, dummy.y-70, 200, 750);

			controller.time.delayedCall(750, ()=> {
				dummy.moveTo(dummy.x+400, dummy.y, 150);
				controller.scene.get("Volume").stopMusic(3500)
            })
            
            controller.time.delayedCall(2750, ()=> {
				scene.tweens.add({
					targets: scene.data.values.backdrop,
					alpha: 0,
					duration: 1500
				});
			});

            controller.time.delayedCall(4250, ()=> {
				scene.data.values.overlay.setVisible(false)
				scene.data.values.actors.objects.resetButton.setInteractive({ useHandCursor: true });
				scene.tweens.add({
					targets: scene.data.values.actors.objects.resetButton,
					alpha: 0.3,
					duration: 2000
				});

                controller.registry.values.cutscene = false;
                stopActiveTextbox(0);
                controller.disableMenu();
			})
			
        }
    },
    cancelEnding: {
        text: [
            ()=>"step1"
        ],
        step1(controller) {
			let scene = controller.scene.get("Forest");
			let dummy = scene.data.values.actors.objects.dummy;
			let player = scene.data.values.actors.player;

			controller.registry.values.cutscene = true;
			scene.data.values.targetDummy = true;
            hideActiveTextbox();

			dummy.swapIn(player);
			scene.cameras.main.startFollow(dummy);
			dummy.moveTo(dummy.x, dummy.y+50, 200, 750);

			controller.time.delayedCall(750, ()=> {
                dummy.swapOut(player)
                scene.cameras.main.startFollow(player)
				scene.data.values.targetDummy = false;
            })
            
            controller.time.delayedCall(850, ()=> {
                stopActiveTextbox();
                controller.registry.values.cutscene = false;
				callTextbox();
			})
			
        }
    }
    
}];

var Preloader = new Phaser.Class({ 
    Extends: Phaser.Scene,
    initialize: function Preloader (){
        Phaser.Scene.call(this, {key: 'Preloader', pack: { files: [ { type: 'image', key: 'moon', url: 'http://ams3.digitaloceanspaces.com/oceanfalls/2018/forestwalkaround/assets/img/moon.png'}]}});
    },

    init: function (data){
        this.cameras.main        
            .setSize(650, 450)
            .setBackgroundColor(0x172f49);
        
        //Scenes to start the game with
        this.cameras.main.once('camerafadeoutcomplete', event => {
            this.scene.start('Controller');
        });

        this.input.on('pointerdown', () => {
            clickOffCanvas = false;
        });
    },

    preload: function (){
        var moon = this.add
            .image(190, 40, "moon")
            .setOrigin(0)
            .setScale(9);

        this.tweens.add({
            targets: moon,
            y: { value: '-=5' },
            duration: 700,
            yoyo: true,
            ease: 'Linear',
            repeat: -1
         });

        var percentBox = this.add.graphics();
        var percentBar = this.add.graphics();
        percentBox
            .fillStyle(0x1f3f61)
            .fillRect(190, 335, 270, 100);
        percentBar
            .fillStyle(0x172f49)
            .fillRect(200, 345, 0, 80);

        var percentText = this.make.text({
            x: 325,
            y: 375,
            text: '0%',
            style: {
                font: '20px monospace',
                fill: '#ffffff'
            }
        });
        percentText.setOrigin(0.5, 0);

        if (!this.sys.game.device.os.desktop){                
            percentBar.destroy();
            percentBox.destroy();
            percentText.setText('This game does not support mobile.');
        }
        else{
            this.load.on('progress', function (value) {
                percentBar.fillRect(200, 345, value * 250, 80);
                percentText.setText(parseInt(value * 99) + '%');
            });
     
            this.load.on('complete', function() {
                this.scene.initiateSprites();
                percentText.destroy();
                percentBox.destroy();
            });

            url = "http://ams3.digitaloceanspaces.com/oceanfalls/2018/forestwalkaround/assets/";

            this.load.bitmapFont("prstart", url+"img/prstart.png", url+"img/prstart.fnt");
            this.load.atlasXML("atlas", url+'img/sprites.png', url+'img/sprites.xml');
            this.load.image("forestBG", url+"img/forestBG.png");
            this.load.tilemapTiledJSON("forestMap", url+"img/forestMap.json");

            this.load.audio("bgm", [url+"audio/bgm.ogg", url+"audio/bgm.mp3",]);
            this.load.audioSprite("sfx", url+"audio/sfx.json", [url+"audio/sfx.ogg", url+"audio/sfx.mp3"]);
            this.load.audioSprite("ambient", url+"audio/ambient.json", [url+"audio/ambient.ogg", url+"audio/ambient.mp3"]);
        }
    },

    create: function (){
        var hitArea = new Phaser.Geom.Rectangle(190, 335, 270, 100);

        var startButton = this.add.graphics();
        var startFill = this.add.graphics();
        var startHover = this.add.graphics();       
        var text = this.add
            .bitmapText(247, 371, 'prstart', "BEGIN")
            .setScale(2); 

        startButton
            .fillStyle(0x1f3f61)
            .fillRect(190, 335, 270, 100)
            .setInteractive( hitArea, Phaser.Geom.Rectangle.Contains)
            .on('pointerover', () => {
                this.input.setDefaultCursor('pointer');
                startHover.setVisible(true);
                text.setTintFill(0x000000);
            })     
            .on('pointerout', () => {
                this.input.setDefaultCursor('default');
                startHover.setVisible(false);
                text.clearTint();
            })
            .once('pointerup', () => {
                this.input.setDefaultCursor('default');
                this.cameras.main.fadeOut(200);
            });       
        startFill
            .fillStyle(0x172f49)
            .fillRect(200, 345, 250, 80);
        startHover
            .fillStyle(0xfff17c)
            .fillRect(200, 345, 250, 80)
            .setVisible(false);
    },

    initiateSprites: function(){     
        let gv = this.registry.values;
        gv.cDatas = {};
        let configs = [];
        
        //Textbox + UI Icons
        {
            this.textures.addSpriteSheetFromAtlas("ui_volume",{
                atlas: 'atlas',
                frame: 'ui_volume',
                frameWidth: 20,
                frameHeight: 20,
            });
            this.textures.addSpriteSheetFromAtlas("ui_textbox_skip",{
                atlas: 'atlas',
                frame: 'ui_textbox_skip',
                frameWidth: 25,
                frameHeight: 25,
            });
        }

        //Backpack + Menu
        {              	
            this.textures.addSpriteSheetFromAtlas("ui_bag",{
                atlas: 'atlas',
                frame: 'ui_bag',
                frameWidth: 38,
                frameHeight: 38,
            });  

            let packAnimSpeed = 15
            configs.push({
                key: 'packOpen',
                frames: this.anims.generateFrameNumbers("ui_bag", {frames: [6, 7, 8, 9, 10, 12]}),
                frameRate: packAnimSpeed,
                repeat: 0,
            });   	
            configs.push({
                key: 'packClose',
                frames: this.anims.generateFrameNumbers("ui_bag", {frames: [10, 10, 9, 8, 7, 6]}),
                frameRate: packAnimSpeed,
                repeat: 0,
            });
        }

        //Lantern
        {
            this.textures.addSpriteSheetFromAtlas("item_lantern",{
                atlas: 'atlas',
                frame: 'item_lantern',
                frameWidth: 26,
                frameHeight: 26,
            });  
            configs.push({
                key: 'item_lantern_anim',
                frames: this.anims.generateFrameNumbers("item_lantern", {frames: [0, 1]}),
                frameRate: 4,
                repeat: -1,
            });

            this.textures.addSpriteSheetFromAtlas("tooltip_lantern",{
                atlas: 'atlas',
                frame: 'tooltip_lantern',
                frameWidth: 189,
                frameHeight: 187,
            });  
            configs.push({
                key: 'tooltip_lantern_anim',
                frames: this.anims.generateFrameNumbers("tooltip_lantern", {frames: [0, 1]}),
                frameRate: 4,
                repeat: -1,
            });
        }

        //Nino
        {
            this.textures.addSpriteSheetFromAtlas("chara_nino",{
                atlas: 'atlas',
                frame: 'nino',
                frameWidth: 37,
                frameHeight: 48,
            });

            configs.push({
                key: 'chara_nino' + 'down',
                frames: this.anims.generateFrameNumbers('chara_nino', {frames: [1, 0, 2, 0]}),
                frameRate: 5,
                repeat: -1
            });
            configs.push({
                key: 'chara_nino' + 'left',
                frames: this.anims.generateFrameNumbers('chara_nino', {frames: [4, 3, 5, 3]}),
                frameRate: 5,
                repeat: -1
            });
            configs.push({
                key: 'chara_nino' + 'right',
                frames: this.anims.generateFrameNumbers('chara_nino', {frames: [7, 6, 8, 6]}),
                frameRate: 5,
                repeat: -1
            });
            configs.push({
                key: 'chara_nino' + 'up',
                frames: this.anims.generateFrameNumbers('chara_nino', {frames: [10, 9, 11, 9]}),
                frameRate: 5,
                repeat: -1
            });

            gv.cDatas.pData = {
                key: "player",
                textureKey: "chara_nino",
                spawnKey: "spawn",
                initialDirection: "left",
                sizeX: 22, 
                sizeY: 12,
                offsetX: 6,
                offsetY: 36,
                speed: 175,
                stillframes: {down:0, left:3, right:6, up:9}
            };
        }

        //Nino Dark
        {
            this.textures.addSpriteSheetFromAtlas("nino_dark",{
                atlas: 'atlas',
                frame: 'nino_dark',
                frameWidth: 37,
                frameHeight: 48,
            });

            configs.push({
                key: 'nino_dark' + 'down',
                frames: this.anims.generateFrameNumbers('nino_dark', {frames: [1, 0, 2, 0]}),
                frameRate: 5,
                repeat: -1
            });
            configs.push({
                key: 'nino_dark' + 'left',
                frames: this.anims.generateFrameNumbers('nino_dark', {frames: [4, 3, 5, 3]}),
                frameRate: 5,
                repeat: -1
            });
            configs.push({
                key: 'nino_dark' + 'right',
                frames: this.anims.generateFrameNumbers('nino_dark', {frames: [7, 6, 8, 6]}),
                frameRate: 5,
                repeat: -1
            });
            configs.push({
                key: 'nino_dark' + 'up',
                frames: this.anims.generateFrameNumbers('nino_dark', {frames: [10, 9, 11, 9]}),
                frameRate: 5,
                repeat: -1
            });
        }


        configs.forEach(config => this.anims.create(config));
    }
});

var Menu = new Phaser.Class({
	Extends: Phaser.Scene,

    initialize: function Menu (){
        Phaser.Scene.call(this, 'Menu');
    },  

    init: function (){
        this.cameras.main
            .setOrigin(0)
            .setRoundPixels(true);
        this.input.on('pointerdown', () => {
            clickOffCanvas = false;
        });
	},

    preload: function (){

    },

    create: function (){

        this.scene.run('Tooltip'); 
        this.data.values.tooltipScene = this.scene.get("Tooltip");

        this.cameras.main.x = -40;

        this.data.values.enabled = false;
        this.data.values.selection = 0;
        this.data.values.selected = 0;
        this.data.values.cooldown = false;

        this.data.values.backpack = this.physics.add
            .sprite(-26, 48, 'ui_bag', 0)
            .setAlpha(0.5)
            .setScale(2)
            .setDepth(5)
        this.data.values.bagPit = this.physics.add
            .image(-22, 12, 'atlas', 'ui_bag_pit')
            .setOrigin(0);

        this.data.values.cursor = this.physics.add
            .image(0, 49, 'atlas', 'ui_selector_box')
            .setScale(2)
            .setDepth(5)
            .setTint(this.registry.values.party[0].colour)
            .setAlpha(0);

        this.data.values.pitMask = this.physics.add
            .image(48, 6, '')
            .setDisplaySize(650, 450)
            .setOrigin(0)
            .setVisible(false)
            .createBitmapMask();
        this.data.values.pit = this.physics.add
            .image(48, 6, '')
            .setOrigin(0)
            .setAlpha(0)
            .setMask(this.data.values.pitMask);
        this.data.values.itemBoxes = this.add.group();
        for (let i = 0; i < 6; i++){
            this.data.values.itemBoxes.add(
                this.physics.add.image(0, 0, 'atlas', "ui_item_box").setDepth(1).setAlpha(0).setMask(this.data.values.pitMask)
            );
        }

        this.data.values.itemSprites = this.add.group();
        for (let i = 0; i < 6; i++){
            this.data.values.itemSprites.add(
                this.physics.add.sprite().setDepth(2).setScale(2).setAlpha(0).setMask(this.data.values.pitMask)
            );
        }

        this.registry.values.keys.A.on("down", event => {
            if (this.registry.values.uiState.menuState == "inv") this.itemUpdate("A");
        });

        this.registry.values.keys.D.on("down", event => {
            if (this.registry.values.uiState.menuState == "inv") this.itemUpdate("D");
        });

        this.registry.values.keys.SPACE.on("down", event => {
            if (this.registry.values.uiState.menuState == "inv") this.itemUpdate("SPACE");
        });

		this.registry.values.keys.E.on("down", event => {
            if (this.data.values.enabled && !this.data.values.cooldown) {
                this.data.values.cooldown = true
                this.time.delayedCall(60, ()=> {this.data.values.cooldown = false})
                switch(this.registry.values.uiState.menuState){
                    case "inv":
                        this.invOut();
                        break
                    default:
                        this.invIn();
                }
            }
        });
    },

    update: function (){
    },

    disable: function(){
        if (this.registry.values.uiState.menuState) this.invOut();
        this.data.values.enabled = false;
        this.tweens.add({
            targets: this.cameras.main,
            x: -30,
            duration: 40
        });
    },

    enable: function(){
        if (!this.registry.values.activeArea || this.registry.values.activeArea == "Shop") return;
        this.data.values.enabled = true;
        this.tweens.add({
            targets: this.cameras.main,
            x: 0,
            duration: 40
        });
    },

    invIn: function() {
        this.tweens.killTweensOf(this.data.values.itemBoxes.getChildren().concat(this.data.values.pit, this.data.values.cursor));
        this.registry.values.uiState.menuState = "inv";            
        
        this.data.values.selection = 0;
        
        let boxPositions = [125, 197, 279, 352, 424, 496];

        this.tweens.add({
            targets: this.data.values.backpack,
            x: {start: -26, to: 48},
            alpha: {start: 0.5, to: 1},
            duration: 60
        });
        this.data.values.cursor.y = 49;
        this.tweens.add({
            targets: this.data.values.cursor,
            x: {start: boxPositions[0]-74, to: boxPositions[0]},
            alpha: {start: 0, to: 1},
            duration: 60
        });
        this.tweens.add({
            targets: this.data.values.bagPit,
            alpha: {start: 1, to: 0},
            duration: 60
        });

        boxPositions.forEach((xPos, i) => {
            this.data.values.itemBoxes.getChildren()[i]
                .setPosition(xPos, 49)                                
                .setDepth(1)
                .setVisible(i < boxPositions.length);

            let item = this.registry.values.items[this.registry.values.party[0].inv[i]];

            if (!!item) {
                let texture = item.icon 

                this.data.values.itemSprites.getChildren()[i]
                .setPosition(xPos, 49)
                .setDepth(2)
                .setTexture("atlas", texture)
                .setVisible(this.registry.values.party[0].inv[i]);

                if (item.icon_anim) {
                    this.data.values.itemSprites.getChildren()[i].anims.play(item.icon_anim);
                }
                else {
                    this.data.values.itemSprites.getChildren()[i].anims.stop();
                }
            }
        });
        
        this.data.values.pit
            .setTexture("atlas", "ui_inventory_pit")
            .setX(48-400);
        this.data.values.itemBoxes.incX(-400);
        this.data.values.itemSprites.incX(-400);

        this.tweens.add({
            targets: this.data.values.itemBoxes.getChildren().concat(this.data.values.pit, this.data.values.itemSprites.getChildren()),
            x: "+=400",
            alpha: {start: 0, to: 1},
            duration: 60
        })
        
        
        this.data.values.tooltipScene.open(boxPositions[0], 90, this.registry.values.items[this.registry.values.party[0].inv[this.data.values.selection]]);
        this.sound.playAudioSprite("sfx", "bagOpen");
        this.data.values.backpack.anims.play("packOpen");
    },

    invOut: function() {
        this.tweens.killTweensOf(this.data.values.itemBoxes.getChildren().concat(this.data.values.pit, this.data.values.cursor));
        this.data.values.tooltipScene.close();
        this.registry.values.uiState.sorting = false;

        this.tweens.killTweensOf(this.data.values.cursor);
        this.registry.values.uiState.menuState = undefined;
        
        this.data.values.selection = 0;

        this.tweens.add({
            targets: this.data.values.itemBoxes.getChildren().concat(this.data.values.pit, this.data.values.itemSprites.getChildren(), this.data.values.cursor),
            x: "-=400",
            alpha: {start: 1, to: 0},
            duration: 60,
            onComplete: () => {this.data.values.itemBoxes.setVisible(false)}
        })

        this.data.values.backpack.anims.play("packClose");
        this.tweens.add({
            targets: this.data.values.backpack,
            x: {start: 48, to: -26},
            alpha: {start: 1, to: 0.5},
            duration: 60
        });
        this.tweens.add({
            targets: this.data.values.bagPit,
            alpha: {start: 0, to: 1},
            duration: 60
        });
         
        this.sound.playAudioSprite("sfx", "bagClose");
    },

    itemUpdate: function(key){
        let boxPositions = [125, 197, 279, 352, 424, 496];        
        let sprites = this.data.values.itemSprites.getChildren();

        if (key == "A" || key == "D") {
            let prev = this.data.values.selection; 

            if (key == "A") {
                this.data.values.selection--;
                if (this.data.values.selection < 0){
                    this.data.values.selection = boxPositions.length - 1;
                }

                this.tweens.add({
                    targets: this.data.values.backpack,
                    angle: {start: 0, to: -5}, 
                    yoyo: true,
                    duration: 40
                })
            }
            else{
                this.data.values.selection++;
                if (this.data.values.selection >= boxPositions.length){
                    this.data.values.selection = 0;
                }

                this.tweens.add({
                    targets: this.data.values.backpack,
                    angle: {start: 0, to: 5}, 
                    yoyo: true,
                    duration: 40
                })                
            }
                    
            if (this.registry.values.uiState.sorting){
                this.tweens.add({
                    targets: [this.data.values.itemBoxes.getChildren()[prev], sprites[prev]],
                    x: boxPositions[prev], 
                    duration: 70
                })
                this.tweens.add({
                    targets: [this.data.values.itemBoxes.getChildren()[this.data.values.selection], sprites[this.data.values.selection]],
                    x: boxPositions[this.data.values.selected], 
                    duration: 70
                })
                this.tweens.add({
                    targets: [this.data.values.cursor, this.data.values.itemBoxes.getChildren()[this.data.values.selected], sprites[this.data.values.selected]],
                    x: boxPositions[this.data.values.selection], 
                    duration: 70
                })

                this.sound.playAudioSprite("sfx", "menu4");
            }
            else {
                this.tweens.add({
                    targets: this.data.values.cursor,
                    x: boxPositions[this.data.values.selection], 
                    duration: 70
                })
                
                let tooltipImage = this.registry.values.items[this.registry.values.party[0].inv[this.data.values.selection]];

                this.data.values.tooltipScene.moveTo(boxPositions[this.data.values.selection], 90, tooltipImage)

                this.sound.playAudioSprite("sfx", "menu2");
            }
        }
        else if (key == "SPACE"){ // && !this.tweens.isTweening(this.data.values.cursor)
            this.registry.values.uiState.sorting = !this.registry.values.uiState.sorting;
            if (this.registry.values.uiState.sorting){
                this.data.values.selected = this.data.values.selection;
                this.data.values.itemBoxes.getChildren()[this.data.values.selected].setDepth(3)
                
                if (sprites.length > 0) {
                    sprites[this.data.values.selected].setDepth(4);
                }

                this.tweens.add({
                    targets: [this.data.values.cursor, this.data.values.itemBoxes.getChildren()[this.data.values.selected], sprites[this.data.values.selected]],
                    y: 79, 
                    duration: 70
                });

                this.data.values.tooltipScene.close();
                this.sound.playAudioSprite("sfx", "menu3");
            }
            else{
                
                this.data.values.itemBoxes.getChildren()[this.data.values.selected].setDepth(1);                 
                if (sprites.length > 0) {
                    sprites[this.data.values.selected].setDepth(2);
                }
                
                this.tweens.add({
                    targets: [this.data.values.cursor, this.data.values.itemBoxes.getChildren()[this.data.values.selected], sprites[this.data.values.selected]],
                    x: boxPositions[this.data.values.selection],
                    y: 49, 
                    duration: 70
                });

                
                this.sound.playAudioSprite("sfx", "menu5");

                [this.data.values.itemBoxes.getChildren()[this.data.values.selected], this.data.values.itemBoxes.getChildren()[this.data.values.selection]]
                = [this.data.values.itemBoxes.getChildren()[this.data.values.selection], this.data.values.itemBoxes.getChildren()[this.data.values.selected]];
                
                [sprites[this.data.values.selected], sprites[this.data.values.selection]]
                = [sprites[this.data.values.selection], sprites[this.data.values.selected]];
                

                [this.registry.values.party[0].inv[this.data.values.selected], this.registry.values.party[0].inv[this.data.values.selection]]
                = [this.registry.values.party[0].inv[this.data.values.selection], this.registry.values.party[0].inv[this.data.values.selected]]
                
                this.scene.get('Controller').inventoryUpdate(this.registry.values.party[0])


                let tooltipImage = this.registry.values.items[this.registry.values.party[0].inv[this.data.values.selection]];
                this.data.values.tooltipScene.open(boxPositions[this.data.values.selection], 90, tooltipImage);
            }
        }

    },
});

var Text = new Phaser.Class({
	Extends: Phaser.Scene,

    initialize: function Text (){
        Phaser.Scene.call(this, 'Text');
    },

    init: function (data){
        this.data.values.scriptVars = data.scriptVars;
		this.cameras.main   
            .setViewport(22, 0, 606, 134)
            .setOrigin(0)
            .setVisible(false);

        this.input.on('pointerdown', () => {
            clickOffCanvas = false;
        });
	},

    preload: function (){

    },

    create: function (){
        this.data.values.textLines = [] //Four lines total
        for (let i = 0; i < 4; i++){
            this.data.values.textLines.push(this.add.group())
        }

        this.clearText();
    },    

    clearText: function(){
        this.data.values.textObjectCount = 0;
        this.data.values.scrollCount = 0;
        this.data.values.scrollTotal = 0;
        this.data.values.scrollTimer = 0;
        this.data.values.scrollCoords = {x: 0, y:0};
        this.data.values.effects = {
            tint: undefined,
            font: undefined,
            anim: undefined,
            scroll: undefined,
        }
        this.data.values.textLines.forEach(textGroup => {
            textGroup.clear(true, true);
        });
    },

    update: function (){
		if (this.data.values.scriptVars.scrolling){
			this.updateScroll();
		}
    },

    newTextObject: function(text, yIndex) {
        let x = 24;
        this.data.values.textLines[yIndex].children.entries.forEach(text => {
            x += text.data.values.text.length * 16;
        });
        
        let y = 23 + (yIndex * 24);

        let textObject = this.add
            .bitmapText(x, y, "prstart", "")
            .setScrollFactor(0)
            .setOrigin(0)
            .setData("text", text)
            .setData("scrollConfig", {speed: 2, timer: 0});
        if (this.data.values.effects.tint){
            textObject.setTintFill(this.data.values.effects.tint);
        }
        if (this.data.values.effects.font){
            textObject.setFont(this.data.values.effects.font);
        }
        if (this.data.values.effects.scroll){
            textObject.setData("scrollConfig", this.data.values.effects.scroll);
        }

        return textObject;
    },

    display: function(text, doScroll = true){        
        this.clearText();
        let filteredText = text.split(/([$#%&]\[\w*]|\n)/g).filter(Boolean)

        let tempRow = 0;
        for (let i = 0; i < filteredText.length; i++){

            let snippet = filteredText[i]            
            let textEffect = snippet.match(/([#$%&])\[(\w*)]/);

            if (textEffect) { //Apply colours, font changes, scroll speed, etc
                this.applyEffect(textEffect[1], textEffect[2]);      
            }
            else if (/\n/.test(snippet)) { //Apply Linebreak   
                tempRow++;
                if (tempRow > 3) break;
            }           
            else { //Print text as normal
                this.data.values.textLines[tempRow].add(this.newTextObject(snippet, tempRow));            
            }                       
        };

        this.data.values.textLines.forEach(textGroup => {
            this.data.values.textObjectCount  += textGroup.getLength();
        });

        this.data.values.scriptVars.scrolling = doScroll;
        if (!doScroll) this.endScroll();
    },

    applyEffect: function(mode = "", param = ""){
        switch(mode){
            case "#": //Hex colour
                let validHex = !(/^(f)\1*$/i.test(param) || /[^0-9a-f]/i.test(param));
                this.data.values.effects.tint = validHex ? parseInt("0x" + param) : undefined;
                break;
            case "$": //New Font
                this.data.values.effects.font = this.getFont(param);
                break;
            case "%": //Cool Effects
                break;
            case "&": //Scroll speed
                if (param == "") param = 3;
                let scrollSpeed = Math.min(Math.max(parseInt(param), 1), 3) - 1;
                let speeds = [1, 1, 1]
                let timers = [7, 2, 0]
                this.data.values.effects.scroll = {
                    speed: speeds[scrollSpeed], timer: timers[scrollSpeed]
                }
                break;
        }
    },

    getFont: function(fontParam){
        switch (fontParam){
            case ("i" || "italic" || "italics"): return "prstartitalic";
            default: return undefined;
        }  
    },

    endScroll: function(){
        this.data.values.scriptVars.scrolling = false;
        this.data.values.scrollCount = 0;
        this.data.values.scrollTotal = 0;
        this.data.values.scrollCoords = {x: 0, y:0};
        this.data.values.textLines.forEach(textGroup => {
            textGroup.children.entries.forEach(textChild => {
                textChild.setText(textChild.data.values.text)
            });
        });
    },
    
    updateScroll: function (){
        let currentTextGroup = this.data.values.textLines[this.data.values.scrollCoords.y]
        let currentTextObject = currentTextGroup.children.entries[this.data.values.scrollCoords.x]

        if (this.data.values.scrollTimer > 0){
            this.data.values.scrollTimer -= 1;
        }
        else{
            this.data.values.scrollTimer = currentTextObject.data.values.scrollConfig.timer;            

            let temp = this.data.values.scrollCount; //First character to go in
            this.data.values.scrollTotal += 1;
            this.data.values.scrollCount += currentTextObject.data.values.scrollConfig.speed; //Character after last Character to go in
            currentTextObject.text += currentTextObject.data.values.text.slice(temp, this.data.values.scrollCount);


            if (!this.scene.get("Textbox").data.values.skippingDialogue && /\w/.test(currentTextObject.text[currentTextObject.text.length-1])) {
                if (currentTextObject.data.values.scrollConfig.speed == 2 && !(this.data.values.scrollTotal % 2 == 0)){
                }  
                else{
                    this.sound.playAudioSprite("sfx", "talknoise");
                }
            }


        }

        //Once the current text object has scrolled all its text, run some checks to move on to the next one
        if (currentTextObject.text.length >= currentTextObject.data.values.text.length) {
            this.data.values.scrollCount = 0;

            this.data.values.scrollCoords.x++;
            if (this.data.values.scrollCoords.x >= currentTextGroup.children.entries.length || this.groupWidth(currentTextGroup) > 576) {
                this.data.values.scrollCoords.x = 0;
                this.data.values.scrollCoords.y++;
            }

            this.data.values.textObjectCount--;
            this.data.values.scriptVars.scrolling = !(this.data.values.textObjectCount == 0 || this.data.values.scrollCoords.y > 3);

            if (!this.data.values.scriptVars.scrolling) this.data.values.scrollCoords = {x:0, y:0};
        }
    },    

    drawChoicetext: function (choiceText, choiceSelection){        
        let text = "";
        for (let i = 0; i < choiceText.length; i++){	
            if (typeof(choiceText[i][0]) === "string"){
                if (choiceText[i].length == 2){
                    if (choiceSelection == i){
                        text += "> ";
                    }
                    else{
                        text += "  ";
                    }
                }
                text += (choiceText[i][0] + "\n");
            }
        }
        this.display(text, false);
    },

    groupWidth: function(group){
        let width = 0
        group.children.entries.forEach(text => {
            width += text.width;
        });
        return width;
    }
});

var Textbox = new Phaser.Class({
	Extends: Phaser.Scene,

    initialize: function Textbox (){
        Phaser.Scene.call(this, 'Textbox');
    },  

    init: function (){
        this.cameras.main
            .setVisible(false);

        this.input.on('pointerdown', () => {
            clickOffCanvas = false;
        });
	},

    preload: function (){

    },

    create: function (){

        //Initializes animation timeline, textbox and text        
        this.data.values.textboxScreenTop = false;
        this.data.values.skippingDialogue = false;
        this.data.values.skipDelay = 0;

        this.data.values.scriptVars = {
            current: undefined,
            count: 0,

            choiceIsActive: false,
            choiceSelection: 0,

            talksprite: undefined,

            scrolling: false,
        };	      
        this.scene.run('Text', {scriptVars: this.data.values.scriptVars}); 
        this.data.values.textScene = this.scene.get("Text");

        this.data.values.script = {
            fallback:{
                text: ["Dialogue not found."],
            }
        };
        globalScript.forEach( script => {
            Object.assign(this.data.values.script, script);
        });
        // globalScript = undefined;

        this.data.values.pointer = this.input.activePointer;

		this.data.values.textboxTimeline = this.tweens.createTimeline();
		this.data.values.textbox = this.add
            .image(22, 0, "atlas", 'ui_textbox')
            .setOrigin(0)		
            .setScrollFactor(0)            
			.setInteractive({ useHandCursor: true  } )
			.on('pointerdown', event => {
				if (this.registry.values.uiState.textbox && !this.data.values.scriptVars.choiceIsActive && !this.registry.values.cutscene) {
                    this.data.values.skippingDialogue = false;
					this.textHandler(undefined);	
				}
            });
            
		this.data.values.talksprite = this.add
            .image(0, 0, "")
            .setScrollFactor(0)
            .setOrigin(0)
            .setVisible(false);
        this.data.values.textboxSkip = this.add
            .sprite(615, 0, "ui_textbox_skip")	
            .setScrollFactor(0)            
            .setInteractive({ useHandCursor: true  } )
			.on('pointerdown', event => {
                this.data.values.skippingDialogue = !this.data.values.skippingDialogue;
            });
            
        this.data.values.talksprite.setDepth(3);
		this.data.values.textbox.setDepth(2);
        this.data.values.textboxSkip.setDepth(1);
    },

    update: function (){
        if (this.registry.values.uiState.textbox && Phaser.Input.Keyboard.JustDown(this.registry.values.keys.SPACE) && !this.registry.values.cutscene){
            if (this.data.values.scriptVars.choiceIsActive) {
                this.selectChoicetext();
            }
            else{
                this.data.values.skippingDialogue = false;
                this.textHandler();
            }
        }
        
        if (this.data.values.scriptVars.choiceIsActive) this.updateChoicetext();

        this.data.values.textboxSkip.setVisible(!this.registry.values.choiceIsActive);
        if (this.data.values.scriptVars.choiceIsActive) this.data.values.skippingDialogue = false;
        
        if (this.data.values.skippingDialogue) {
            this.data.values.textboxSkip.setTexture("ui_textbox_skip", 1);
            if (this.data.values.skipDelay++ > 3 && this.data.values.scriptVars.current){
                this.textHandler();
                this.data.values.skipDelay = 0;
            }
        }
        else {
            this.data.values.textboxSkip.setTexture("ui_textbox_skip", 0); 
        }

    },

    textHandler: function (key = undefined) {    
        if (this.data.values.scriptVars.scrolling) {
            this.data.values.textScene.endScroll();
            return;
        }
        if (this.registry.values.cutscene){
            return;
        }
        
        //Basic validation to make sure the provided key exists in the script
        //Only goes through the process if one isn't already in use to ensure the key doesn't change halfway through some dialogue
        if (!this.data.values.scriptVars.current){

            if ((typeof(key) !== "string") || !(key in this.data.values.script)){
                console.warn(`Key (${key}) is invalid. Resorting to fallback.`);
                key = "fallback";
            }	
            
            this.registry.values.flags.add("read_"+key);
            this.data.values.scriptVars.current = this.data.values.script[key];
        }

        //If the current key doesn't exist, just dont open the box at all
        //if (!this.data.values.textboxTimeline.isPlaying() && this.data.values.scriptVars.current){  
        if (!this.data.values.textboxTimeline.isPlaying()){            
            let currentText = this.data.values.scriptVars.current.text[this.data.values.scriptVars.count]

            //If it's a string, it's just dialogue
            if (typeof(currentText) === "string") {		

                //If you can find a @[] in the string, there's a talksprite to be cut out             
                let splitString = currentText.split(/^@\[(?:(\d+)([a-z]+)|([a-z]+)(\d+))\]/i).filter(Boolean)
                if (splitString.length == 3){
                    currentText = splitString[2];

                    let flipped = /\d/.test(splitString[0]); // Testing for ["n", 1] or [1, "n"]
                    let texture = flipped ? [splitString[1], splitString[0]] : [splitString[0], splitString[1]]
                                       
                    this.data.values.scriptVars.talksprite = {
                        texture: texture,
                        flipped: flipped
                    } ;
                }
                else{
                    this.data.values.scriptVars.talksprite = undefined;	
                }

                this.data.values.textScene.display(currentText);
                this.data.values.scriptVars.count++;
            }
            //If it's an array, it's describing a choice. 
            else if (Array.isArray(currentText)){
                this.data.values.scriptVars.choiceIsActive = true;
                this.data.values.scriptVars.talksprite = undefined;
                
                //Aims to start cursor on a valid option
                while (currentText[this.data.values.scriptVars.choiceSelection].length < 2){
                    this.data.values.scriptVars.choiceSelection++;
                }

                this.data.values.textScene.drawChoicetext(currentText, this.data.values.scriptVars.choiceSelection);
                
            }  
            //If it's a function, it'll return the name of a method to be run
            else if (typeof(currentText) === "function") {
                //this.data.values.scriptVars.talksprite = undefined;
                this.data.values.scriptVars.count++;
                
                let actionKey = currentText();
                if (actionKey in this.data.values.scriptVars.current){
                    this.data.values.scriptVars.current[actionKey](this.scene.get("Controller"));
                    // Make sure to put a callTextbox() in your function if you want it to progress text
                }
                else{
                    //Fallback in case method isn't found
                    this.textHandler();
                }
            }  
            //If currentText is undefined, we've presumably hit the end of the dialogue
            else if (!currentText){
                this.data.values.scriptVars.current = undefined;
                this.textboxDown();
            }
            
            this.updateTalksprite();
    
            //Textbox will only tween up if it's inactive, and the text hasn't just been wiped.
            if (!this.registry.values.uiState.textbox && this.data.values.scriptVars.current){
                //Use up just-down buffer to prevent a keypress bleeding through from outside the textbox
                Phaser.Input.Keyboard.JustDown(this.registry.values.keys.S)
                Phaser.Input.Keyboard.JustDown(this.registry.values.keys.W)
                this.registry.values.uiState.textbox = true;
                this.textboxUp();
            }
        }
    },
    
    updateChoicetext: function (){
        var currentText = this.data.values.scriptVars.current.text[this.data.values.scriptVars.count];
        if (Phaser.Input.Keyboard.JustDown(this.registry.values.keys.S)) {
            do{
                this.data.values.scriptVars.choiceSelection++;
                if (this.data.values.scriptVars.choiceSelection >= currentText.length){
                    this.data.values.scriptVars.choiceSelection = 0;
                }
            } while (currentText[this.data.values.scriptVars.choiceSelection].length == 1);
            this.data.values.textScene.drawChoicetext(currentText, this.data.values.scriptVars.choiceSelection);
        }
        else if (Phaser.Input.Keyboard.JustDown(this.registry.values.keys.W)) {
            do{
                this.data.values.scriptVars.choiceSelection--;
                if (this.data.values.scriptVars.choiceSelection < 0){
                    this.data.values.scriptVars.choiceSelection = currentText.length-1;
                }
            } while (currentText[this.data.values.scriptVars.choiceSelection].length == 1);
            this.data.values.textScene.drawChoicetext(currentText, this.data.values.scriptVars.choiceSelection);
        }
    },

    selectChoicetext: function() {
        this.data.values.scriptVars.choiceIsActive = false;

        var currentText = this.data.values.scriptVars.current.text[this.data.values.scriptVars.count];
        let newKey = currentText[this.data.values.scriptVars.choiceSelection][1];        
        this.data.values.scriptVars.choiceSelection = 0;

        if (newKey.length == 0) {
            //stopActiveTextbox(100);
            this.data.values.scriptVars.count++;
            this.textHandler();
        }
        else{
            switchDialogueKey(newKey);
        }
    
    },
    
    updateTalksprite: function (){
        if (this.data.values.scriptVars.talksprite){
            this.data.values.talksprite.setTexture(this.data.values.scriptVars.talksprite.texture[0], this.data.values.scriptVars.talksprite.texture[1]);
            this.data.values.talksprite.visible = true;
    
            if (this.data.values.scriptVars.talksprite.flipped){
                this.data.values.talksprite.scaleX = -1;
                this.data.values.talksprite.x = 611;
                this.data.values.textScene.cameras.main.x = 22;
            }
            else{
                this.data.values.talksprite.scaleX = 1;
                this.data.values.talksprite.x = 39;
                this.data.values.textScene.cameras.main.x = 121;
            }
        }
        else{
            this.data.values.talksprite.visible = false;
            this.data.values.textScene.cameras.main.x = 22;
        }
    },
    
    textboxUp: function (animDuration = 100){
        this.scene.get("Controller").disableMenu();

        let bounceDuration = 0.3*animDuration;

        this.cameras.main.setVisible(true);
        this.data.values.textScene.cameras.main.setVisible(true);

        this.updateTalksprite();
    
        //Tweens break if duration is 0, so just skip it entirely
        if (animDuration <= 0){
            this.data.values.textbox.y = this.data.values.textScene.cameras.main.y = 303;            
            this.data.values.talksprite.y = 320;
            this.data.values.textboxSkip.y = 287;
        }
        else{
            if (this.data.values.textboxScreenTop) {
                this.data.values.textbox.y = this.data.values.textScene.cameras.main.y = 13 - 200;            
                this.data.values.talksprite.y = 30 - 200;
                this.data.values.textboxSkip.y = 162 - 200;
            }
            else{
                this.data.values.textbox.y = this.data.values.textScene.cameras.main.y = 303 + 200;            
                this.data.values.talksprite.y = 320 + 200;
                this.data.values.textboxSkip.y = 287 + 200;
            }

            let signs = this.data.values.textboxScreenTop ? ['+', '-'] : ['-', '+'];
    
            this.data.values.textboxTimeline = this.tweens.timeline({	
                targets: [this.data.values.textbox, this.data.values.textboxSkip, this.data.values.talksprite, this.data.values.textScene.cameras.main],
                tweens: [{ 
                    y: { value: signs[0]+'=215' },
                    duration: animDuration,
                },
                { 	y: { value: signs[1]+'=15' },
                    duration: bounceDuration,		
                }]
            });
        }
    
    },
    
    textboxDown: function (remainActive = false, animDuration = 100){
        let bounceDuration = 0.3*animDuration;

        //Tweens break if duration is 0, so just skip it entirely
        if (animDuration <= 0){
            this.data.values.textbox.y = this.data.values.textScene.cameras.main.y = 303 + 200;            
            this.data.values.talksprite.y = 320 + 200;
            this.data.values.textboxSkip.y = 287 + 200;
            this.textboxDownCallback(undefined, undefined, this, remainActive);
        }
        else {
            if (this.data.values.textboxScreenTop) {
                this.data.values.textbox.y = this.data.values.textScene.cameras.main.y = 13;            
                this.data.values.talksprite.y = 30;
                this.data.values.textboxSkip.y = 162;
            }
            else{
                this.data.values.textbox.y = this.data.values.textScene.cameras.main.y = 303;            
                this.data.values.talksprite.y = 320;
                this.data.values.textboxSkip.y = 287;
            }
            
            let signs = this.data.values.textboxScreenTop ? ['+', '-'] : ['-', '+'];

            this.data.values.textboxTimeline = this.tweens.timeline({
                targets: [this.data.values.textbox, this.data.values.textboxSkip, this.data.values.talksprite, this.data.values.textScene.cameras.main],
                tweens:[{
                    y: { value: signs[0]+'=15' },
                    duration: bounceDuration,
                },
                {	y: { value: signs[1]+'=215' },
                    duration: animDuration,
                    onComplete: this.textboxDownCallback,
                    onCompleteParams: [this, remainActive],
                }]
            });
        }
    },
    
    textboxDownCallback: function (a, b, textboxScene, remainActive){
        textboxScene.registry.values.uiState.textbox = remainActive;
        if (!remainActive){
            textboxScene.data.values.scriptVars.current = undefined
            textboxScene.data.values.scriptVars.count = 0

            textboxScene.data.values.skippingDialogue = false;
            textboxScene.data.values.skipDelay = 0;
    
            textboxScene.data.values.scriptVars.choiceIsActive = false
            textboxScene.data.values.scriptVars.choiceSelection = 0
    
            textboxScene.data.values.scriptVars.talksprite = undefined
    
            textboxScene.data.values.scriptVars.scrolling = false

            textboxScene.updateTalksprite();
            textboxScene.data.values.textScene.clearText();
            
            textboxScene.scene.get("Controller").enableMenu();
        }
        textboxScene.cameras.main.setVisible(false);
        textboxScene.data.values.textScene.cameras.main.setVisible(false);
    },

    screenTopToggle: function (animDuration = 100){
        //let bounceDuration = animDuration * 0.3;
        
        if (this.data.values.textboxTimeline.isPlaying()) return;

        if (this.data.values.textboxScreenTop) {
            this.data.values.textbox.y = this.data.values.textScene.cameras.main.y = 13;            
            this.data.values.talksprite.y = 30;
            this.data.values.textboxSkip.y = 162;
        }
        else{
            this.data.values.textbox.y = this.data.values.textScene.cameras.main.y = 303;            
            this.data.values.talksprite.y = 320;
            this.data.values.textboxSkip.y = 287;
        }

        let signs = this.data.values.textboxScreenTop ? ['+', '-'] : ['-', '+'];
        this.data.values.textboxScreenTop = !this.data.values.textboxScreenTop;

        if (this.registry.values.uiState.textbox){
            this.data.values.textboxTimeline = this.tweens.timeline({
                tweens: [
                // { 	y: { value: signs[1]+'=15' },
                //     duration: bounceDuration,
                //     targets:[this.data.values.textbox, this.data.values.textboxSkip, this.data.values.talksprite, this.data.values.textScene.cameras.main]	
                // },
                {   y: { value: signs[0]+'=125' },
                    duration: animDuration,
                    targets: this.data.values.textboxSkip
                },
                {   y: { value: signs[0]+'=290' },
                    duration: animDuration,
                    offset: 0,
                    targets:[this.data.values.textbox, this.data.values.talksprite, this.data.values.textScene.cameras.main]
                },
                // { 	y: { value: signs[1]+'=15' },
                //     duration: bounceDuration,	
                //     offset: animDuration + bounceDuration,
                //     targets:[this.data.values.textbox, this.data.values.textboxSkip, this.data.values.talksprite, this.data.values.textScene.cameras.main]	
                // }
            ]
            });
        }
    },

});

var Tooltip = new Phaser.Class({
	Extends: Phaser.Scene,

    initialize: function Tooltip (){
        Phaser.Scene.call(this, 'Tooltip');
    },  

    init: function (){
        this.cameras.main
            .setOrigin(0)
            .setRoundPixels(true);
        this.input.on('pointerdown', () => {
            clickOffCanvas = false;
        });
	},

    preload: function (){

    },

    create: function (){
        this.data.values.maskShape = this.add
            .sprite(0, 0, '')
            .setOrigin(0, 1)
            .setVisible(false);
        this.data.values.mask = this.data.values.maskShape.createBitmapMask();

        this.data.values.tooltip = this.physics.add
            .sprite(0, 0)
            .setOrigin(0, 0)
            .setDepth(1)    
            .setVisible(false)                  
            .setMask(this.data.values.mask);

        this.data.values.footer = this.physics.add
            .image(0, 0, 'atlas', 'ui_tooltip_footer')
            .setOrigin(0, 1)
            .setDepth(3)
            .setVisible(false);

        this.data.values.itemSprites = this.add.group();
        let iconPositions = [
            {x: 34, y: 270},
            {x: 74, y: 270},
            {x: 114, y: 231},
            {x: 154, y: 231},
            {x: 114, y: 270},
            {x: 154, y: 270},
        ];
        for (let i = 0; i < 6; i++){
            this.data.values.itemSprites.add(
                this.physics.add
                    .image(iconPositions[i].x, iconPositions[i].y)
                    .setDepth(2)
                    .setVisible(false)                  
                    .setMask(this.data.values.mask)
            );
        }

    },

    update: function (){
    },

    displayInventory: function (inventory) {
        for (let i = 0; i < 6; i++){
            let texture = this.registry.values.items[inventory[i]];
            texture = (!!texture) ? texture.icon : undefined;

            this.data.values.itemSprites.getChildren()[i]
                .setTexture('atlas', texture)
                .setVisible(!!inventory[i]);
        }

    },

    open: function(xPos, yPos, item){
        if (this.data.values.tooltip.visible){
            this.moveTo(xPos, yPos, item);
            return;
        }
        if (!item) return;

        let textureKey = item.tooltip;

        this.data.values.tooltip
            .setTexture('atlas', textureKey)
            .setVisible(true);
        
        if (item.tooltip_anim) {
            this.data.values.tooltip.anims.play(item.tooltip_anim);
        }
        else {
            this.data.values.tooltip.anims.stop();
        }

        this.data.values.footer
            .setPosition(0, 0)
            .setVisible(true)
            .setScale(1);

        this.data.values.maskShape
            .setPosition(0, 0)
            .setDisplaySize(this.data.values.tooltip.width, this.data.values.tooltip.height);

        this.cameras.main.setPosition(xPos - (this.data.values.tooltip.width/2), yPos)
        
        if ("inv" in item) this.displayInventory(item.inv);
        else this.data.values.itemSprites.setVisible(false);

        let startPos = 2 * this.data.values.footer.height;
        let endPos = this.data.values.tooltip.height;

        this.tweens.add({
            targets: [this.data.values.footer, this.data.values.maskShape],
            y: {start: startPos, to: endPos},
            duration: 70,
            onComplete: () => {
                this.data.values.footer.setVisible(false);
            }
        });
    },

    moveTo: function(xPos, yPos, item){
        if (!this.data.values.tooltip.visible){
            this.open(xPos, yPos, item);
            return;
        }
        if (!item){
            this.close();
            return
        }

        let textureKey = item.tooltip;
        
        this.data.values.tooltip.setTexture('atlas', textureKey);
        if (item.tooltip_anim) {
            this.data.values.tooltip.anims.play(item.tooltip_anim);
        }
        else {
            this.data.values.tooltip.anims.stop();
        }
        if ("inv" in item) this.displayInventory(item.inv);
        else this.data.values.itemSprites.setVisible(false);

        this.tweens.add({
            targets: this.cameras.main,
            x: xPos - (this.data.values.tooltip.width/2),
            y: yPos,
            duration: 70,
            onComplete: () => {
                if (item.tooltip_anim) {
                    this.data.values.tooltip.anims.play(item.tooltip_anim);
                }
                else {
                    this.data.values.tooltip.anims.stop();
                }
            }
        });

    },
    
    close: function (){
        if (!this.data.values.tooltip.visible){
            return;
        }
        this.data.values.maskShape
            .setPosition(0, this.data.values.tooltip.height);
        this.data.values.footer
            .setPosition(0, this.data.values.tooltip.y)            
            .setVisible(true)
            .setScale(1, -1);
        this.tweens.add({
            targets: [this.data.values.footer, this.data.values.maskShape],
            y: `+=${this.data.values.tooltip.height-(this.data.values.footer.height*2)}`,
            duration: 70
        }).on('complete',() => {
            this.data.values.tooltip.anims.stop();
            this.data.values.tooltip.setVisible(false);
            this.data.values.footer.setVisible(false);
            this.data.values.itemSprites.setVisible(false);
        })
    },
});

var Volume = new Phaser.Class({
	Extends: Phaser.Scene,

    initialize: function Volume (){
        Phaser.Scene.call(this, 'Volume');
    },

    init: function (){
		this.cameras.main        
            .setSize(650, 450);	

        this.input.on('pointerdown', () => {
            clickOffCanvas = false;
        });
	},

    preload: function (){

    },

    create: function (){
        this.sound.setVolume(this.registry.values.globalVolume);

        this.volControl = this.physics.add
            .sprite(640, 10, "ui_volume", this.registry.values.muteStage)
            .setOrigin(1, 0)
            .setScale(2)
            .setAlpha(.4)
            .setInteractive( { useHandCursor: true  } )                
            .on('pointerover', event => {  
                this.tweens.add({
                    targets: this.volControl,
                    alpha: 0.8,
                    duration: 60
                });
            })    
            .on('pointerout', event => {    
                this.tweens.add({
                    targets: this.volControl,
                    alpha: 0.4,
                    duration: 60
                });    
            })     
            .on('pointerdown', event => {
                this.registry.values.muteStage += (this.registry.values.muteStage >= 2) ? -2 : 1;
                this.volControl.setTexture("ui_volume", this.registry.values.muteStage);
                if ('overlay' in this.registry.values.music) this.registry.values.music.overlay.setSeek(this.registry.values.music.seek)
                this.sound.setVolume(this.registry.values.globalVolume * [1, 0.3, 0][this.registry.values.muteStage]);                
            });
    },

    update: function (){
    },

    playTrack: function (key, playOverlay = true){
        if (!key in this.registry.values.tracks){
            console.warn(`No music with key "${key}" exists`);
        }
        else {
            if (this.registry.values.music) this.stopMusic();
            this.registry.values.music = this.registry.values.tracks[key];
            this.registry.values.music.setVolume(1);
            this.registry.values.music.play();
            if ("overlay" in this.registry.values.music) {
                this.registry.values.music.on('looped', () => { 
                    this.registry.values.music.overlay.setSeek(this.registry.values.music.seek)
                })
                this.registry.values.music.overlay.setVolume(playOverlay? 1 : 0).play();
            }
        }
    },    

    stopMusic: function(fadeDuration = 0){
        if (!this.registry.values.music.key) {
            console.warn('No music is currently selected.')
        }
        else {
            this.tweens.killTweensOf(this.registry.values.music)
            if (fadeDuration > 0) {
                let targets = [this.registry.values.music]
                if (this.registry.values.music.overlay) targets.push(this.registry.values.music.overlay)
                this.tweens.add({
                    targets:  targets,
                    volume:   0,
                    duration: fadeDuration,
                });
            } else {
                if ("overlay" in this.registry.values.music) {
                    this.registry.values.music.off('looped');
                    this.registry.values.music.overlay.stop();
                }
                this.registry.values.music.stop()
            }
        }
    },

    fadeOutOverlay: function(fadeDuration = 250){
        if (!("overlay" in this.registry.values.music)){
            // console.warn(`Current track "${this.registry.values.music.key}" either has no overlay, or does not exist.`);
        }
        else {
            this.registry.values.music.overlay.setSeek(this.registry.values.music.seek)
            this.tweens.add({
                targets:  this.registry.values.music.overlay,
                volume:   0,
                duration: fadeDuration,
            });
        }
    },

    fadeInOverlay: function(fadeDuration = 250){
        if (!("overlay" in this.registry.values.music)){
            // console.warn(`Current track "${this.registry.values.music.key}" either has no overlay, or does not exist.`);
        }
        else {
            this.tweens.add({
                targets:  this.registry.values.music.overlay,
                volume:   1,
                duration: fadeDuration,
            });
        }
    },

    setMusicVolume(level = undefined, fadeDuration = 1, includeOverlay = true) {
        if (!Number.isFinite(level)) {
            console.warn("New volume is not finite")
            return
        }

        let targets = [this.registry.values.music]
        if (includeOverlay && "overlay" in this.registry.values.music) targets.push(this.registry.values.music.overlay)
        targets.forEach(target => {
            this.tweens.killTweensOf(target)
        });
        
        this.tweens.add({
            targets:  targets,
            volume:   level,
            duration: fadeDuration,
        });
    },

    fadeToTrack: function(key, fadeDuration = 250, playOverlay = true){
        if (!key in this.registry.values.tracks){
            console.warn(`No music with key "${key}" exists`)
        }
        else {		
            targets = [this.registry.values.music];
            if ("overlay" in this.registry.values.music) targets.push(this.registry.values.music.overlay);
            this.tweens.add({
                targets:  targets,
                volume:   0,
                duration: fadeDuration,
                onComplete: () => {			
                    this.time.delayedCall(fadeDuration*0.3, () =>{
                        this.registry.values.music.stop();
                        this.registry.values.music.setVolume(1);
                        if ("overlay" in this.registry.values.music) {
                            this.registry.values.music.overlay.setVolume(1);
                            this.registry.values.music.overlay.stop();
                        }
                        this.playTrack(key, playOverlay);
                    });	
                    
                }
            });
        }
    }
});


class gameActorList {
    constructor(scene){
        this.scene = scene;
        this.objects = {};
        this.npcs = scene.physics.add.group();
		this.depthList = [];
    }

    update(){
        for (let obj in this.objects){
            this.objects[obj].update();
        } 
        this.depthSort();
    }

    newPlayer(cData){
        cData.key = "player";
        return this.push(cData.key, Player.new(this.scene, cData))
    }
    newFollower(cData){
        let target = cData.target in this.objects ? this.objects[cData.target] : null;
        return this.push(cData.key, Follower.new(this.scene, cData, target))
    }
    newNPC(cData){
        //creates a copy of the NPC in its own object. It helps to have them seperate.
        let chara = NPC.new(this.scene, cData);
        this.npcs.add(chara);

        //Adding npc to a group resets this for some godforsaken reason
        chara.setImmovable(true);

        return this.push(cData.key, chara);
    }
    newCutsceneActor(cData){
        return this.push(cData.key, CutsceneActor.new(this.scene, cData));
    }

    push(key, object, excludeFromDepthSort = false){
        this.objects[key] = object;
        if (!excludeFromDepthSort) this.depthList.push(key);
        return object;
    }

    depthSort(){
        //Uses insertion sort to sort sprite depths
        for (var i = 0, length = this.depthList.length; i < length; i++) {
            let temp = this.depthList[i];
            for (var j = i - 1; j > -1 && this.objects[this.depthList[j]].y > this.objects[temp].y; j--) {
                this.depthList[j + 1] = this.depthList[j];
            }
            this.depthList[j + 1] = temp;
        };

        let depth = 1;
        this.depthList.forEach((key, index) => {
            this.objects[key].setDepth(depth);
            depth++;
        });
    }

    get player(){
        if ("player" in this.objects)
            return this.objects.player;
        else{
            console.warn("No character named 'player' currently exists");
            return null;
        }
    }
}

class gameChara extends Phaser.Physics.Arcade.Sprite {
    /*Character data is object containing (* indications optional):
        scene: Scene object that the character belongs to
        key: String. Unique ID for character.
        textureKey: String. Reference to spritesheet used by character.
        spawnKey: String
        * sizeX: Int. Width of underlying object beneath sprite.
        * sizeY: Int. Height of underlying object beneath sprite.
        * offsetX: Int. X offset of underlying object beneath sprite.
        * offsetY: Int. T offset of underlying object beneath sprite.
        * interactSize: Int. Determines length of a player's interaction box. Defaults to 15.
        * direction: Optional string. Direction character is facing. Defaults to "down".
        * speed: Optional int. Speed character moves. Defaults to 200.
        * target: String, only for followers. Defaults to "player".
    */
    
    constructor(scene, characterData, bodyType){
        super(scene, 0, 0, characterData.textureKey)
        scene.sys.displayList.add(this);
        scene.sys.updateList.add(this);        
        scene.sys.arcadePhysics.world.enableBody(this, bodyType !== 1 ? 0 : 1);

        this.scene = scene;
        this.cData = characterData;
        this.name = this.cData.key;

        if (!("speed" in this.cData)) this.cData.speed = 200;
        if (!("interactSize" in this.cData)) this.cData.interactSize = 15;
        if (!("direction" in this.cData)) this.cData.direction = "down";
        if ("initialDirection" in this.cData) this.cData.direction = this.cData.initialDirection;

        this.setSpawn();
        
        //Character position is taken from their center
	    //We want to position them by their feet, so the spawnpoint is slightly offset
        this.setPosition(this.cData.spawnX, this.cData.spawnY-(this.height/2))
            //.setTexture(this.cData.textureKey, 0);

        if ("sizeX" in this.cData && "sizeY" in this.cData)
            this.setSize(this.cData.sizeX, this.cData.sizeY);
        if ("offsetX" in this.cData && "offsetY" in this.cData)  
            this.setOffset(this.cData.offsetX, this.cData.offsetY);
    }

    setSpawn() {
        if (typeof(this.cData.spawnKey) === "string"){
            let spawnPoint = this.scene.data.values.map.findObject("SpawnLayer", obj => obj.name === this.cData.spawnKey);
            if (spawnPoint != null){
                this.cData.spawnX = spawnPoint.x;
                this.cData.spawnY = spawnPoint.y;
                if (spawnPoint.properties != null){
                    for (let i = 0;  i < spawnPoint.properties.length; i++){
                        if (spawnPoint.properties[i].name == "direction"){
                            this.cData.direction = spawnPoint.properties[i].value;
                            break;
                        }
                    }
                }
            }		
        }
    }

    teleport(){
        this.setSpawn()
    
        //Character position is taken from their center
        //We want to position them by their feet, so the spawnpoint is slightly offset
        this.setPosition(this.cData.spawnX, this.cData.spawnY-(this.height/2));
        this.stopAnim();
    }

    stopAnim(){
        this.anims.stop();
        // If we were moving, pick an idle frame to use
        this.setTexture(this.cData.textureKey, this.cData.stillframes[this.cData.direction]);
    }

    update(){
    }

    isMoving(){
        return (Math.abs(this.body.deltaXFinal()) > 0.1 || Math.abs(this.body.deltaYFinal()) > 0.1);
    }

    destroy(){
        super.destroy();
    }

    interactHandler(){

    }
}

class Player extends gameChara {
    constructor(scene, characterData) {
        super(scene, characterData, 0);

        this.scene.data.values.pInteractBox = this.scene.physics.add
			.image(0, 0, "")
			.setVisible(false);

		//Initialize camera to follow player
		this.scene.cameras.main.startFollow(this);

		//Adds permanent overlap checks for flags
        this.scene.physics.world.addOverlap(this, this.scene.data.values.flags, this.scene.flagHandler, null, this.scene);
    }

    static new(scene, characterData) {
        let charaIsValid = (
            "key" in characterData &&
            "textureKey" in characterData &&
            "spawnKey" in characterData
        );
        if (!charaIsValid){
            console.warn("Unable to create player with given params");
            return null;
        }
        else
            return new Player(scene, characterData);
    }

    update(){
        super.update();
        // Stop any previous movement from the last frame	
        //this.body.setVelocity(0);	

        if (!this.scene.registry.values.uiState.uiIsActive && !this.scene.registry.values.transitioning){
            //Sets directional velocity depending on buttons held
            let xDir = 0;
            let yDir = 0;
            if (this.scene.registry.values.keys.A.isDown) xDir--;
            if (this.scene.registry.values.keys.D.isDown) xDir++;
            if (this.scene.registry.values.keys.W.isDown) yDir--;
            if (this.scene.registry.values.keys.S.isDown) yDir++;	
            
            this.body.setVelocity(xDir * this.cData.speed, yDir * this.cData.speed);	

            // Normalize and scale the velocity so that the character can't move faster along diagonals
            if (this.body.blocked.none){
                this.body.velocity
                    .normalize()
                    .scale(this.cData.speed);
            }
            //Picks animation for unimpeded movement. Prioritises horizontal animation over vertical
            //deltaX/Y threshold is 0.1 instead of 0 to prevent floating point collision errors from messing with the animation
            if (this.isMoving()){
                if 		(this.body.deltaXFinal() < -0.1) {this.cData.direction = "left";} 
                else if (this.body.deltaXFinal() >  0.1) {this.cData.direction = "right";} 
                else if (this.body.deltaYFinal() < -0.1) {this.cData.direction = "up";} 
                else if (this.body.deltaYFinal() >  0.1) {this.cData.direction = "down";}
                this.anims.play(this.cData.textureKey+this.cData.direction, true);	
            }
            //Allows you to change direction to face wall when currently blocked by one
            else if (!this.body.touching.none){
                if 		(xDir < 0 && yDir == 0 && this.body.touching.left ) {this.cData.direction = "left";}
                else if (xDir > 0 && yDir == 0 && this.body.touching.right) {this.cData.direction = "right";}
                else if (yDir < 0 && xDir == 0 && this.body.touching.up   ) {this.cData.direction = "up";}
                else if (yDir > 0 && xDir == 0 && this.body.touching.down ) {this.cData.direction = "down";}
                this.stopAnim();
            }
            //Stops animation if no direction is given
            if (xDir == 0 && yDir == 0){
                this.stopAnim();
            }
        }
        else{
            //Forces character to stop while textbox is active
            this.body.setVelocity(0);
            this.stopAnim();
        }

        this.updateInteractBox();
    }

    updateInteractBox(){
        //Scales interaction box to the size and position of the player's hitbox
        let interactX = this.body.x + this.body.halfWidth,
            interactY = this.body.y + this.body.halfHeight,
            interactWidth = this.body.width,
            interactHeight = this.body.height;
    
        switch (this.cData.direction){
            case "left":
                interactX -= Math.round(this.cData.interactSize/2) - 1;
                //interactY -= this.body.height;
                interactWidth += this.cData.interactSize;
                //interactHeight += this.body.height;
                break;
            case "right":
                interactX += Math.round(this.cData.interactSize/2);
                //interactY -= this.body.height;
                interactWidth += this.cData.interactSize;
                //interactHeight += this.body.height;
                break;
            case "up":
                //interactX = 0;
                interactY -= Math.round(this.cData.interactSize/2);
                //interactWidth = 0;
                interactHeight += this.cData.interactSize;
                break;
            case "down":
                //interactX = 0;
                interactY += Math.round(this.cData.interactSize/2) - 1;
                //interactWidth = 0;
                interactHeight += this.cData.interactSize;
                break;
        }
    
        this.scene.data.values.pInteractBox
            .setPosition(interactX, interactY)
            .setSize(interactWidth, interactHeight);
    }

    destroy(){
        super.destroy();
        if (this.scene) this.scene.data.values.pInteractBox.destroy();
    }
}

class Follower extends gameChara {
    constructor(scene, characterData, target) {
        super(scene, characterData, 0);
        this.target = target;
        this.cData.queue = [];
        this.cData.queueMax = 10;
    }

    static new(scene, characterData, target) {
        let charaIsValid = (
            "key" in characterData &&
            "textureKey" in characterData &&
            "spawnKey" in characterData &&
            "target" in characterData &&
            target !== null
        );
        if (!charaIsValid){
            console.warn("Unable to create follower with given params");
            return null;
        }
        else
            return new Follower(scene, characterData, target);
    }

    update(){
        super.update();
        if (!this.scene.registry.values.uiState.uiIsActive && this.target.isMoving()){

           
            
            //If follower is too far away, speed up a bit
            let currentSpeed = this.scene.data.values.actors.player.cData.speed;
            if (Phaser.Math.Distance.Between(this.x, this.y, this.target.x, this.target.y) > 50) {
                currentSpeed += 150;
            }
            
            //Extends range character follows at on high refresh rate monitors
            if (this.scene.game.loop.delta > 12){
                if (this.cData.queueMax == 24) this.cData.queue = []
                this.cData.queueMax = 10;
            }
            else this.cData.queueMax = 24;

            //this.cData.queueMax = Math.floor(-1.4286714*this.scene.game.loop.delta) + 34

            //Notes the player's current coordinates
            this.cData.queue.push({
                x: this.target.x,
                y: this.target.y,
            });

            if (this.cData.queue.length > this.cData.queueMax){
                //Removes the oldest coordinate from the queue and moves towards it
                this.scene.physics.moveToObject(this, this.cData.queue.shift(), currentSpeed);
            }
            else if (Phaser.Math.Distance.Between(this.x, this.y, this.target.x, this.target.y) > 40) {
                //If queue doesn't have enough coords prepared, move directly towards target instead
                this.scene.physics.moveToObject(this, this.target, currentSpeed);
            }
    
            // Update the animation last and give left/right animations precedence over up/down animations
            if (this.isMoving()) {
                if 		(this.body.deltaXFinal() < -0.1) {this.cData.direction = "left";} 
                else if (this.body.deltaXFinal() >  0.1) {this.cData.direction = "right";} 
                else if (this.body.deltaYFinal() < -0.1) {this.cData.direction = "up";} 
                else if (this.body.deltaYFinal() >  0.1) {this.cData.direction = "down";}
                this.anims.play(this.cData.textureKey+this.cData.direction, true);	
            } 
            else {
                this.stopAnim();
            }
        }
        else {
            //Forces character to stop while player is still, or UI is active.
            this.body.setVelocity(0);	
            this.stopAnim();
        }
    }

    teleport(){
        this.cData.queue = [];
        super.teleport();
    }
}

class NPC extends gameChara {
    constructor(scene, characterData) {
        super(scene, characterData, 0);
        this.prevDirection = this.cData.direction;
    }

    static new(scene, characterData) {
        let charaIsValid = (
            characterData &&
            "key" in characterData &&
            "textureKey" in characterData
        );
        if (!charaIsValid) {
            console.warn("Unable to create npc with given params");
            console.log(characterData)
            return null;
        }
        else
            return new NPC(scene, characterData);
    }

    update() {
        super.update();
        let player = this.scene.data.values.actors.player;

        //Random wandering?

        if (this.isMoving()) {
            if 		(this.body.deltaXFinal() < -0.1) {this.cData.direction = "left";} 
            else if (this.body.deltaXFinal() >  0.1) {this.cData.direction = "right";} 
            else if (this.body.deltaYFinal() < -0.1) {this.cData.direction = "up";} 
            else if (this.body.deltaYFinal() >  0.1) {this.cData.direction = "down";}
            this.anims.play(this.cData.textureKey+this.cData.direction, true);	
        } 
        else {
            this.stopAnim();
        }
        
        this.prevDirection = this.cData.direction;
    }

    interactHandler(player){
        super.interactHandler();
        if (this.scene.registry.values.uiState.textbox){
            let isLeft = (this.body.left >= player.body.right);
            let isRight = (this.body.right <= player.body.left);
            let isAbove = (this.body.top >= player.body.bottom);
            let isBelow = (this.body.bottom <= player.body.top)
            if 		(isLeft) {this.cData.direction = "left";} 
            else if (isRight) {this.cData.direction = "right";} 
            else if (isAbove) {this.cData.direction = "up";} 
            else if (isBelow) {this.cData.direction = "down";}
        }
    }
}

class CutsceneActor extends gameChara {
    constructor(scene, characterData) {
        super(scene, characterData, 0);
        this.cData.queue = [];
        this.cData.autoDirection = true;
        this.prevDirection = this.cData.direction;
    }

    static new(scene, characterData) {
        let charaIsValid = (
            characterData &&
            "key" in characterData &&
            "textureKey" in characterData &&
            "spawnKey" in characterData
        );
        if (!charaIsValid) {
            console.warn("Unable to create cutscene actor with given params");
            console.log(characterData)
            return null;
        }
        else
            return new CutsceneActor(scene, characterData);
    }

    update() {
        super.update();

        if (this.isMoving()) {
            if (this.cData.queue.length > 0) {
                let distance = Phaser.Math.Distance.Between(this.x, this.y, this.cData.queue[0].x, this.cData.queue[0].y);
                if (distance < (this.body.speed / 75)) {
                    this.body.reset(this.cData.queue[0].x, this.cData.queue[0].y);
                    this.cData.queue.shift();
                    if (this.cData.queue.length > 0) {
                        this.scene.physics.moveTo(this, this.cData.queue[0].x, this.cData.queue[0].y, this.cData.queue[0].speed, this.cData.queue[0].maxTime)
                    }
                }
            }

            if (this.cData.autoDirection) {
                if 		(this.body.deltaXFinal() < -0.1) {this.cData.direction = "left";} 
                else if (this.body.deltaXFinal() >  0.1) {this.cData.direction = "right";} 
                else if (this.body.deltaYFinal() < -0.1) {this.cData.direction = "up";} 
                else if (this.body.deltaYFinal() >  0.1) {this.cData.direction = "down";}
            }
            this.anims.play(this.cData.textureKey+this.cData.direction, true);	
        } 
        else {
            this.stopAnim();
        }
        
        this.prevDirection = this.cData.direction;
    }

    swapIn(chara){
        this.setPosition(chara.x, chara.y)
            .setVisible(true)
        this.cData.direction = chara.cData.direction
        this.cData.textureKey = chara.cData.textureKey
        chara.setVisible(false)
    }
    swapOut(chara){
        chara.setPosition(this.x, this.y)
            .setVisible(true)
        chara.cData.direction = this.cData.direction
        chara.cData.textureKey = this.cData.textureKey
        this.setVisible(false)
    }
    moveTo(x, y, speed = 60, maxTime = 0){
        if (this.cData.queue.length === 0) {
            this.scene.physics.moveTo(this, x, y, speed, maxTime)
        }
        this.cData.queue.push({x, y, speed, maxTime})
    }
}

function callTextbox(key = undefined){
	game.scene.getScene("Textbox").textHandler(key);
}

function switchDialogueKey (key){
	let scene = game.scene.getScene("Textbox")
	if (scene.data.values.scriptVars.current){
		scene.data.values.scriptVars.count = 0;
	}
	scene.data.values.scriptVars.current = undefined;
	callTextbox(key);
}

function stopActiveTextbox(time = 100){
	game.scene.getScene("Textbox").textboxDown(false, time);
}

function hideActiveTextbox(){
	game.scene.getScene("Textbox").textboxDown(true);
}

function unhideActiveTextbox(){
	game.scene.getScene("Textbox").textboxUp();
}

function flipTextboxTop(){
	game.scene.getScene("Textbox").screenTopToggle();
}

function setTextboxTop(value = false){
	game.scene.getScene("Textbox").data.values.textboxScreenTop = value;
}


function swapCharacters(chara1, chara2){
	let tempPosition = {
		x: chara1.x,
		y: chara1.y
	};
	let tempData = {}

	chara1.setPosition(chara2.x, chara2.y);
	chara2.setPosition(tempPosition.x, tempPosition.y);	

	Object.getOwnPropertyNames(chara1.cData).forEach((key) => {
		if (!/^(key|target)$/.test(key)) {
			tempData[key] = chara1.cData[key]
			chara1.cData[key] = undefined;
		}
	});

	Object.getOwnPropertyNames(chara2.cData).forEach((key) => {
		if (!/^(key|target)$/.test(key)) {
			chara1.cData[key] = chara2.cData[key]
			chara2.cData[key] = undefined;
		}
	});

	Object.getOwnPropertyNames(tempData).forEach((key) => {
		chara2.cData[key] = tempData[key]
	});

	[chara1, chara2].forEach((chara) =>{
		chara.setTexture(chara.cData.textureKey)
		if ("sizeX" in chara.cData && "sizeY" in chara.cData){
			chara.setSize(chara.cData.sizeX, chara.cData.sizeY)
		}
		if ("offsetX" in chara.cData && "offsetY" in chara.cData) {
			chara.setOffset(chara.cData.offsetX, chara.cData.offsetY);
		}
		chara.cData.queue = (chara.cData.target) ? [] : undefined;
	})
}

function areaTransition(scene, camera, targetScene, targetSpawnPoint, fadeDuration = 0){
	scene.registry.values.transitionisFade = (fadeDuration > 0);
	scene.registry.values.transitionFadeDuration = fadeDuration;

	scene.registry.values.transitioning = true;

	scene.registry.values.cDatas.pData.spawnKey = targetSpawnPoint;

	if (fadeDuration > 0){
		camera.once('camerafadeoutcomplete', event => {
			scene.registry.values.activeArea = targetScene;
			scene.scene.switch(targetScene);		
		});	
		camera.fadeOut(fadeDuration);
	}
	else{
		scene.registry.values.activeArea = targetScene;
		scene.scene.switch(targetScene);
	}
}

//Creates usable collision objects from shapes on the tilemap.
function objectArrayToImageArray(objects, scene){
	if (objects == null) {
        console.warn('Cannot create images using given array');
        return;
    }
    let images = [];
    for (let i = 0; i < objects.length; i++) {
        let object = objects[i];
       	let image = scene.make.image();
       	image.name = object.name;
       	image.setOrigin(0, 0);
       	image.x = object.x;
		image.y = object.y;
       	image.displayWidth = object.width;
       	image.displayHeight = object.height;
		if (object.rotation){
            let angle = DegToRad(object.rotation);
            Rotate(offset, angle);
            image.rotation = angle;
        }
        image.visible = false;
        for (let key in object.properties){
            image.setData(object.properties[key].name, object.properties[key].value);
        }
		images.push(image); 
    }
    return images;
}

var config = {
	type: Phaser.AUTO,
	parent: "game-container",
	width: 650,
	height: 450,
	scale:{
        autoCenter: Phaser.Scale.CENTER_BOTH,
	},
	pixelArt: true,
	physics: {
		default: "arcade",
		arcade: {gravity: {y: 0}}
	},
	parent: "game-container",
	scene: [ Preloader, Controller ],
}; 

var game = new Phaser.Game(config);

var clickOffCanvas = true;
var containerID = document.getElementById("ht-page");

containerID.addEventListener('mousedown', function(e) {
	game.input.keyboard.enabled = !clickOffCanvas;
    clickOffCanvas = true;
});